/* 
 Routines for hierarchical computations in CoR.
 Copyright (C) 2014, Ayan Chakrabarti <ayanc@ttic.edu>
*/

#include <string.h>
#include "vecops.h"

#define MIN(a,b) ( (a) < (b) ? (a) : (b) )
#define MAX(a,b) ( (a) > (b) ? (a) : (b) )

/* These should exist in whatever this is linked to */
void * env_alloc(size_t bytes);
void env_free(void * ptr);

double * mblock;

/* Go up one step of the pyramid, aggregating from children           */
/* out is parent layer, in is child layer, P is patch size of parents */
void up1(int w,int h,
	int P, double * in, double * out,
	int numt, int tid)
{
	int P2;
	int j,jst,jend;

	P2=P/2;    /* Offset of second child in each dim */

	/* Break up for threading */
	jst = ((w-P+1)*tid)/numt;
	jend = ((w-P+1)*(tid+1))/numt;
	
	for(j = jst; j < jend; j++)
		vec_add4(&out[j*h], 
			 &in[j*h], &in[j*h+P2],
			 &in[(j+P2)*h], &in[(j+P2)*h+P2],
			 (h-P+1));

}

/* Go down one step of the pyramid, aggregating from children           */
/* out is child layer, in is parent layer, P is patch size of children  */
void down1(int w,int h,
	   int P, double * in, double *in0, double * out,
	   int numt, int tid)
{
	int j,jst,jend;

	/* Break up for threading */
	jst = ((w-P+1)*tid)/numt;
	jend = ((w-P+1)*(tid+1))/numt;

	for(j = jst; j < MIN(jend,P);j++) {

		memcpy(&out[j*h],&in[j*h],P*sizeof(double));
		
		vec_add2(&out[j*h+P],
			 &in[j*h],&in[j*h+P],
			 (h-3*P+1));
			 
		
		memcpy(&out[j*h+(h-2*P+1)],&in[j*h+(h-3*P+1)],P*sizeof(double));
	}


	for(j = MAX(P,jst); j < MIN(jend,w-2*P+1);j++) {

		vec_add2(&out[j*h],
			 &in[(j-P)*h],&in[j*h],
			 P);
		
		vec_add4(&out[j*h+P],
			 &in[(j-P)*h],&in[(j-P)*h+P],
			 &in[j*h],&in[j*h+P],
			 (h-3*P+1));
			 
		vec_add2(&out[j*h+(h-2*P+1)],
			 &in[(j-P)*h+(h-3*P+1)], &in[j*h+(h-3*P+1)],
			 P);
	}


	for(j = MAX(w-2*P+1,jst); j < jend;j++) {

		memcpy(&out[j*h],&in[(j-P)*h],P*sizeof(double));
		
		vec_add2(&out[j*h+P],
			 &in[(j-P)*h],&in[(j-P)*h+P],
			 (h-3*P+1));
			 
		
		memcpy(&out[j*h+(h-2*P+1)],&in[(j-P)*h+(h-3*P+1)],P*sizeof(double));
	}

	if(in0 != NULL)
		for(j = jst; j < jend; j++)
			vec_acc(&out[j*h],&in0[j*h],(h-P+1));
}


/* Down1 for uint16        */
void down1SHORT(int w,int h, 
	   int P, uint16_t * in, uint16_t *in0, uint16_t * out,
	   int numt, int tid)
{
	int j,jst,jend;

	/* Break up for threading */
	jst = ((w-P+1)*tid)/numt;
	jend = ((w-P+1)*(tid+1))/numt;

	for(j = jst; j < MIN(jend,P);j++) {

		memcpy(&out[j*h],&in[j*h],P*sizeof(uint16_t));
		
		vec_add2SHORT(&out[j*h+P],
			      &in[j*h],&in[j*h+P],
			      (h-3*P+1));
			 
		
		memcpy(&out[j*h+(h-2*P+1)],&in[j*h+(h-3*P+1)],P*sizeof(uint16_t));
	}


	for(j = MAX(P,jst); j < MIN(jend,w-2*P+1);j++) {

		vec_add2SHORT(&out[j*h],
			      &in[(j-P)*h],&in[j*h],
			      P);
		
		vec_add4SHORT(&out[j*h+P],
			      &in[(j-P)*h],&in[(j-P)*h+P],
			      &in[j*h],&in[j*h+P],
			      (h-3*P+1));
			 
		vec_add2SHORT(&out[j*h+h-2*P+1],
			      &in[(j-P)*h+(h-3*P+1)], &in[j*h+(h-3*P+1)],
			      P);
	}


	for(j = MAX(w-2*P+1,jst); j < jend;j++) {

		memcpy(&out[j*h],&in[(j-P)*h],P*sizeof(uint16_t));
		
		vec_add2SHORT(&out[j*h+P],
			      &in[(j-P)*h],&in[(j-P)*h+P],
			      (h-3*P+1));
			 
		
		memcpy(&out[j*h+(h-2*P+1)],&in[(j-P)*h+(h-3*P+1)],P*sizeof(uint16_t));
	}

	if(in0 != NULL)
		for(j = jst; j < jend; j++)
			vec_accSHORT(&out[j*h],&in0[j*h],(h-P+1));
}


/* out(d2,d1) = in(d1,d2)*/
void transpose(int d1, int d2, double * in, double * out) {

	int i,j;

	for(j = 0; j < d2; j++)
		for(i = 0; i < d1; i++)
			out[i*d2+j] = in[j*d1+i];

}

/* Computes out(n,m) = \sum_d Z(n,d) U(n,d,m)  */
/* Call transpose after! */
void getTheta1(int w, int h, int M, int D,
		   double * out, double * Z, double * U,
		   int numt, int tid) {

	
	int i, d , m,  ist, iend, wh;

	/* Break up for threading */
	wh = w*h; ist = (tid*wh)/numt; iend = ((tid+1)*wh)/numt;
	iend -= ist;

	for(m = 0; m < M; m++) {
		vec_mul( &out[m*wh+ist], &Z[ist], &U[m*wh*D+ist], iend);
		for(d = 1; d < D; d++)
			vec_macc(&out[m*wh+ist], &Z[d*wh+ist], &U[m*wh*D+d*wh+ist], iend);
	}
}


/* Computes ZJ(n,d) = J(n)*Z(n,d) = \sum_{m=1:M} in(n,m) U(n,d,m) */
/* Call transpose before! */
void getZJ(int w, int h, int M, int D,
	   double * ZJ, double * in, double * U, int numt, int tid) {

	int i, d , m,  ist, iend, wh;
	
	/* Break up for threading */
	wh = w*h; ist = (tid*wh)/numt; iend = ((tid+1)*wh)/numt;
	iend -= ist;

	for(d = 0; d < D; d++) {
		vec_mul(&ZJ[d*wh+ist],&in[ist],&U[d*wh+ist], iend);
		for(m = 1; m < M; m++)
			vec_macc(&ZJ[d*wh+ist],&in[ist+m*wh],&U[m*wh*D+d*wh+ist], iend);
	}
}

/* Build \theta, e pyramid from Z, Z^2 */
void upsweep(int w, int h, int M, int D,
	     double * thetae,
	     int minlev, int maxlev,
	     double * Z, double * Z2, double * U,
	     int numt, int tid) {

	double * child, * parent, * swp;
	double * mblock1, * mblock2;
	int lev,P,m;
	
	if(tid == 0) 
		mblock = env_alloc(2*w*h*(M+1)*sizeof(double));

#pragma omp barrier
	mblock1 = mblock; mblock2 = &mblock[w*h*(M+1)];

	getTheta1(w,h,M,D,mblock1,Z,U,numt,tid);
	if(tid == 0) {
		memcpy(&mblock1[w*h*M],Z2,w*h*sizeof(double));
	}
#pragma omp barrier

	lev = 1; P = 2; child = mblock1; parent = mblock2;
	while(lev <= maxlev) {

		if(lev >= minlev)
			parent = &thetae[w*h*(M+1)*(lev-minlev)];

		for(m = 0; m <= M; m++)
			up1(w,h,P,&child[w*h*m],&parent[w*h*m],numt,tid);
#pragma omp barrier

		P*=2; lev++; swp = parent; parent = child; child = swp;
	}

	if(tid == 0)
		env_free(mblock);
}

/* Reconstruct from estimated theta */
void downsweep(int w, int h, int M, int D,
	     double * Z, uint16_t * J,
	     int minlev, int maxlev,
	     double * theta, uint16_t * I, double * U,
	     int numt, int tid) {

	double * child, * parent,  *swp;
	uint16_t * shchild, * shparent, * shswp;
	int lev,P,m;
	
	if(tid == 0) {
		mblock = env_alloc(2*w*h*M*sizeof(double) + 2*w*h*sizeof(uint16_t));
		memcpy(mblock,&theta[(maxlev-minlev)*w*h*M],w*h*M*sizeof(double));
		memcpy((uint16_t *) &mblock[2*w*h*M], &I[(maxlev-minlev)*w*h], w*h*sizeof(uint16_t));
	}

#pragma omp barrier

	parent = mblock; child = &mblock[w*h*M];
	shparent = (uint16_t *) &mblock[2*w*h*M]; shchild = &shparent[w*h];
	P = 1 << maxlev; lev = maxlev;
	
	while(lev > 0) {
		P/=2; lev--;

		if(lev >= minlev) {
			swp = &theta[(lev-minlev)*w*h*M]; shswp = &I[(lev-minlev)*w*h];
			for(m = 0; m < M; m++)
				down1(w,h,P,&parent[w*h*m],&swp[w*h*m],&child[w*h*m],numt,tid);
		} else {
			shswp = NULL;
			for(m = 0; m < M; m++)
				down1(w,h,P,&parent[w*h*m],NULL,&child[w*h*m],numt,tid);
		}

		down1SHORT(w,h,P,shparent,shswp,shchild,numt,tid);
#pragma omp barrier

		swp = parent; parent = child; child = swp;
		shswp = shparent; shparent = shchild; shchild = shswp;
		if(lev == 1)
			shchild = J;
	}	
	
	getZJ(w,h,M,D,Z,parent,U,numt,tid);
#pragma omp barrier

	if(tid == 0)
		env_free(mblock);
}



/*******************************************************/
/* Code for solving L2 costs                           */
/*******************************************************/

/* Call with i > j*/
int compidx(int M, int i, int j) {
	int idx;
	i-=j;	
	for(idx=0; j > 0;j--) {
		idx += M; M--;
	}
	idx+=i;

	return idx;
}

void block_subm3(int w,int h, double * F, double *F1, double *F2, double * F3, int P, int numt, int tid) {

	int j, jst, jend;

	/* Break up for threading */
	jst = ((w-P+1)*tid)/numt;
	jend = ((w-P+1)*(tid+1))/numt;

	for(j = jst; j < jend; j++)
		vec_m3si(&F[j*h],&F1[j*h],&F2[j*h],&F3[j*h],h-P+1);

}

void block_div(int w,int h, double * F, double *F1, int P, int numt, int tid) {

	int j, jst, jend;

	/* Break up for threading */
	jst = ((w-P+1)*tid)/numt;
	jend = ((w-P+1)*(tid+1))/numt;

	for(j = jst; j < jend; j++)
		vec_div(&F[j*h],&F1[j*h],h-P+1);

}

/* Compute C = A + lambda * B */
void block_sumwt(int w, int h, int M,
		 double * F0, double * F1, double * F2, double lambda,
		 int minlev, int maxlev, int numt, int tid) {

	int P,lev,M2,k,wh;
	int j,jst,jend;

	M2 = M + (M*M-M)/2; wh = w*h;

	P = 1 << minlev; lev = minlev;

	for(; lev <= maxlev; lev++) {
		
		jst = ((w-P+1)*tid)/numt;
		jend = ((w-P+1)*(tid+1))/numt;

		for(k = 0; k < M2; k++)
			for(j = jst; j < jend; j++)
				vec_aplb(&F0[j*h+k*wh],&F1[j*h+k*wh],&F2[j*h+k*wh],lambda,h-P+1);
		
		F0 = &F0[w*h*M2]; F1 = &F1[w*h*M2]; F2 = &F2[w*h*M2];
		P *= 2;
	}
}

/* Cholesky Decomposition Code */
void cholesky(int w, int h, int M,
	      double * F0, 
	      int minlev, int maxlev,
	      int numt, int tid) {

	int P,lev,M2;
	int p, q, r, pp, pq, qq, rp, rq, wh; 
	double * F;

	wh = w*h;
	M2 = M + (M*M-M)/2;

	P = 1 << minlev; lev = minlev;

	for(; lev <= maxlev; lev++) {

		F = &F0[w*h*M2*(lev-minlev)];

		for(p = 0; p < M; p++) {

			pp = compidx(M,p,p)*wh;
			for(q = 0; q < p;q++) {
				pq = compidx(M,p,q)*wh;
				qq = compidx(M,q,q)*wh;
				/* F[p,p] -= F[p,q]*F[p,q]*F[q,q] */
				block_subm3(w,h,&F[pp],&F[pq],&F[pq],&F[qq],P,numt,tid);
			}

			for(r = p+1; r < M; r++) {
				rp = compidx(M,r,p)*wh;
				for(q = 0; q < p; q++) {
					rq = compidx(M,r,q)*wh;
					pq = compidx(M,p,q)*wh;
					qq = compidx(M,q,q)*wh;
					/* F[r,p] -= F[r,q]*F[p,q]*D[q,q]  */
					block_subm3(w,h,&F[rp],&F[rq],&F[pq],&F[qq],P,numt,tid);
				}

				block_div(w,h,&F[rp],&F[pp],P,numt,tid);
			}
		}

		P*= 2;
	}
}


void block_subm(int w,int h, double * F, double *F1, double *F2, int P, int numt, int tid) {

	int j, jst, jend;

	/* Break up for threading */
	jst = ((w-P+1)*tid)/numt;
	jend = ((w-P+1)*(tid+1))/numt;

	for(j = jst; j < jend; j++)
		vec_msi(&F[j*h],&F1[j*h],&F2[j*h],h-P+1);

}


/* Minimize a purely l2 cost, where the cholesky decomposition of F_p has been computed. */
void minl2cost(int w, int h, int M,
	       double * thetae, double * thetac, double * F0,
	       int minlev, int maxlev,
	       int numt, int tid) {
	
	int P,lev,M2;
	int p,q,pp,pq,wh;
	double * F, * te, * tc;

	wh = w*h;
	M2 = M + (M*M-M)/2;

	P = 1 << minlev; lev = minlev;

	for(; lev <= maxlev; lev++) {

		F = &F0[w*h*M2*(lev-minlev)];
		te = &thetae[w*h*(M+1)*(lev-minlev)];
		tc = &thetac[w*h*(M+1)*(lev-minlev)];

		/* Forward substitution, and division */
		for(p = 1; p < M; p++) {
			for(q = 0; q < p; q++) {
				pq = compidx(M,p,q)*wh;
				block_subm(w,h,&te[p*wh],&te[q*wh],&F[pq],P,numt,tid);
			}
		}

		/* Division */
		for(p = 0; p < M; p++) {
			pp = compidx(M,p,p)*wh;
			block_div(w,h,&te[p*wh],&F[pp],P,numt,tid);
		}

		/* Backward substitution */
		for(p = M-2; p >= 0; p--) {
			for(q = M-1; q > p; q--) {
				pq = compidx(M,q,p)*wh;
				block_subm(w,h,&te[p*wh],&te[q*wh],&F[pq],P,numt,tid);
			}
		}

		for(p = 0; p < M; p++)
			block_subm(w,h,&te[M*wh],&te[p*wh],&tc[p*wh],P,numt,tid);

		P *= 2;
	}
}


/* Detect outliers and set corresponding theta to 0 */
void findout(int w, int h, int M, 
	     double * theta, uint16_t * I, double * tcost, uint16_t * nout,
	     double * cost, double * imap, double ocost, int P,
	     int numt, int tid) {

	int wh, i, j, m, jst, jend, idx;
	double thresh;

	jst = ((w-P+1)*tid)/numt;
	jend = ((w-P+1)*(tid+1))/numt;

	for(j = jst; j < jend; j++) {
		idx = h*j;
		for(i = 0; i < h-P+1; i++) {
			thresh = imap[idx]*ocost;
			if(cost[idx] >= thresh) {
				I[idx] = 0;
				(*tcost) += thresh;
				(*nout)++;
				for(m = 0; m < M; m++)
					theta[idx+m*w*h] = (double)0;

			} else {
				I[idx] = 1;
				(*tcost) += cost[idx];
			}
			idx++;
		}
	}
}
