% solveConsensus: Performs alternating minimization to
%                 minimize the consensus objective.
%
% [z,num] = amStereo(z_0,w_0,minlev,maxlev,lambda,outcost,imaps)
%
% Solves the consensus framework with a local planar model, with
% data costs defined as w_0(n) * (z(n) - z_0(n))^2 (See
% paper). Uses fully overlapping (i.e., with stride 1) square
% patches of sides 2.^minlev to 2.^maxlev. Outcost = \tau_0. imaps
% is the intensity-based factor described in the paper, and
% computed by getIMaps.m.
%
% Returns consensus disparity map as z, and |J_n| as num.
% 
% Copyright (C) 2014, Ayan Chakrabarti <ayanc@ttic.edu>
function [d,num] = solveConsensus(din,wts,minl,maxl,lambda0,outcost,imapc)

nlev = maxl-minl+1;

MAXITERS=80;  % Total #iterations
REFINEITERS=30; % #iterations to refine after occ correction.

% Schedule for increasing lambda
lambda = lambda0 / 2^18;
ofctr = 2^(3); LUPFREQ = 6;

% Planar model (M=3)
M = 3; M2 = 6;
[U,U2] = getUUTplanar(size(din));

fprintf('Doing precomputation ...');
lmtx = ones(size(wts));
Fl = upsweep(lmtx,lmtx,U2,minl,maxl);
Fw = upsweep(wts,wts,U2,minl,maxl);

% Set up per-region outlier costs
imaps = zeros([size(din) nlev]);
for i = 1:nlev
  P2 = 4^(minl+i-1);
  imaps(1:size(imapc{i},1),1:size(imapc{i},2),i) = imapc{i}*P2; 
end;
wd = wts.*din; wd2 = wd.*din;

d = din;
fprintf(' Done!\n');

the = zeros([size(din) M+1 nlev]); thc = zeros([size(din) M+1 nlev]);
theta = zeros([size(din) M nlev]); I = uint16(zeros([size(din) nlev]));

lold = Inf;

for iters = 1:MAXITERS
  totcost = 0; numout = 0;

  if lold ~= lambda
    F = choleskSum(Fw,Fl,lambda,M,minl,maxl);
    lold = lambda;
  end;

  % Upsweep to compute
  fprintf('\r ^ ');
  ld = lambda.*d; ld2 = ld.*d + wd2; ld = ld + wd;
  upsweep_M(ld, ld2, U,minl,maxl,'the');
    
  fprintf('\r - ');
  % Solve for theta and find corresponding cost
  minl2cost_M(the,F,minl,maxl,'thc');
  [numout,totcost] = findOut_M(thc,imaps,outcost,minl,maxl,'theta','I');
    
  % Downsweep to compute
  fprintf('\r v ');
  [dnew,num] = downsweep(theta,I,U,minl,maxl);
  dnew = dnew ./ double(num);
  dnew(num == 0) = d(num == 0);
  d = dnew;

  % Apply occlusion correction
  if iters == MAXITERS-REFINEITERS
    d = occProp(d,wts > 0);
  end;
  
  fprintf('\r   Iteration %d: Total cost = %e,',iters,totcost);
  fprintf('%8d outliers            ',numout);
  
  % Increase consistency weight
  if mod(iters,LUPFREQ) == 0
    lambda = min(lambda0,lambda*ofctr);
  end;
end;
%%%
fprintf('\n');


% Returns U(n) and U(n)U(n)^T for local planar model.
function [U,U2] = getUUTplanar(sz)

x = [-(sz(2)-1)/2:(sz(2)-1)/2]; 
y = [-(sz(1)-1)/2:(sz(1)-1)/2]; 
[x,y] = meshgrid(x,y);
U = ones([sz(1) sz(2) 1 3]);
U2 = ones([sz(1) sz(2) 1 5]);

U(:,:,1,1) = x; U(:,:,1,2) = y;
U2(:,:,1,1) = x.^2; U2(:,:,1,2) = x.*y; 
U2(:,:,1,3) = x; U2(:,:,1,4) = y.^2; U2(:,:,1,5) = y; 